import java.util.HashMap;

/** Implémentation efficace du calcul du LCA, utilisant les Range Minimum Queries. */
public class EfficientLCA {

	/** Structure de données permettant de répondre aux 'Minimum Range queries' */
	RMQ rmq=null;

	/** Arbre en entrée */
	Tree node;

	/** ce tableau stoque, pour chaque occurrence d'noeud (selon un parcours préfixe de l'arbre), la distance à la racine */
	Data[] dist;

	// champ utile pour la Question 3.1;
	/** table d'hachage qui stoque, pour chaque noeud de l'arbre (identifié par son étiquette), son indice dans le array dist[] */
	HashMap<String,Integer> indices=new HashMap<String,Integer>();

	EfficientLCA(Tree node, int b) {
		this.node=node;
		int n=Tree.size(node); // compute tree size
		this.dist=new Data[2*n-1]; // create the array dist[] storing all node distances

		this.setDistances(); // initialize the arrays dist[] and the table 'indices'
		
		// Le constructeur EfficientLCA(GT node, int b) est à compléter (Question 3.2);
		// il faut initialiser le champ 'rmq'
		//this.rmq=new BruteForceRMQ(this.dist); // not efficient (brute-force computation of RMQ)
		this.rmq=new EfficientRMQ(this.dist, b); // efficient computation of RMQ
	}

	/**
	 * Remplit (récursivement) le tableau 'dist[]': pour chaque occurrence de chaque noeud calcul sa distance à la racine
	 * 
	 * @param node  	noeud racine de l'arbre
	 * @param distance	distance à la racine
	 * @param count  	paramètre auxiliaire: utile pour compter les entrées dans le tableau 'dist'
	 */
	private int setDistancesAux(Tree node, int distance, int count) {
		this.dist[count]=new Data(node.element, distance);
		count++;

		for(Tree child: node.children) {
			count=setDistancesAux(child, distance+1, count);
			this.dist[count]=new Data(node.element, distance);
			count++;
		}
		return count;
	}

	/** Remplit le tableau dist[] et le tableau indices[] */
	void setDistances() {
		//throw new Error("\nMéthode setDistances() à compléter (Question 3.1)");
		
		// first step: initialize the array dist[]
		this.setDistancesAux(this.node, 0, 0); // initialize the array dist[]
		
		// second step: initialize the hash table 'indices'
		int count=0;
		for(Data data: this.dist) {
			if(this.indices.get(data.s)==null) {
				this.indices.put(data.s, count);
			}
			count++;
		}
	}

	/**
	 * Renvoie le LCA des deux noeuds dont les étiquettes sont respectivement 'el1' et 'el2'. <br> <br>
	 * Remarque: on suppose que <br>
	 * -) 'el1' et 'el2' sont stoqués dans l'arbre <br>
	 * -) il n'y a pas de repetitions ('el1' et 'el2' apparaissent exactement une fois dans l'arbre)
	 * 
	 * @param el1  	étiquette du premier noeud n1
	 * @param el2	étiquette du second noeud n2
	 * @return		le LCA de n1 and n2
	 */
	String computeLCA(String el1, String el2) {
		//throw new Error("\nMéthode computeLCA(el1, el2) à compléter (Question 3.2)");

		int index1=this.indices.get(el1);
		int index2=this.indices.get(el2);

		int l=Math.min(index1, index2);
		int r=Math.max(index1, index2);

		Data result=this.rmq.getMin(l, r);

		return result.s;
	}

}
