/**
 * Algorithme efficace pour le calcul des Range Minimum Queries
 */
public class EfficientRMQ implements RMQ {
	
	/** Tableau en entrée */
	Data[] table;
	
	/** taille des blocks */
	int b;
	
	/** Tableau auxiliaire */
	Data[] B;
	
	/**
	 * Initialise les champs de la classe et remplit le tableau auxiliaire 'B'.
	 * 
	 * @param table 	input array
	 * @param b			taille des bloques
	 */
	EfficientRMQ(Data[] table, int b) {
		//throw new Error("\nConstructeur EfficientRMQ(table, b) à compléter (Question 1.2)");
		
		this.table=table;
		this.b=b;
		BruteForceRMQ rmq=new BruteForceRMQ(table);
		
		int n=this.table.length;
		int l=(n/b); // number of blocks
		if(l*this.b<n) // n is not a multiple of b
			l++;
		this.B=new Data[l];
		
		for(int i=0;i<l;i++) { // iterate over all blocks
			int start=b*i; // index of the first element of the i-th block
			int end=start+b;
			end=Math.min(end, n); // for the last block (which could be shorter)
			
			Data min=rmq.getMin(start, end-1);
			this.B[i]=min;
		}
	}
	
	/** Renvoie la taille du tableau */
	public int getSize() {
		return this.table.length;
	}
	
	/** Renvoie le tableau en entrée */
	public Data[] getArray() {
		return table;
	}
	
	/** Renvoie le tableau auxiliaire B */
	Data[] getAuxiliaryArray() {
		return B;
	}
	
	/** Renvoie un élément de valeur minimale qui se trouve dans le sous-tableau T[l, r]. <br>
	 * On suppose que: <tt> 0 <=l<=r< n <tt>.
	 * 
	 * @param l		extremité gauche de l'intervalle (l>=0)
	 * @param r		extremité droite de l'intervalle (l<=r<n)
	 */
	public Data getMin(int l, int r) {
		if(l<0 || l>r || r>=table.length)
			throw new Error("Error: wrong interval");
		
		//throw new Error("\nméthode getMin(l, r) à compléter (Question 1.3)");
		
		Data result;
		BruteForceRMQ rmqT=new BruteForceRMQ(table);
		BruteForceRMQ rmqB=new BruteForceRMQ(B);
		
		int first=l/this.b; // first block
		int last=r/this.b; // last block
		
		if(first==last || first+1==last) // same block or consecutive blocks
			result=rmqT.getMin(l, r);
		else {
			int endFirst=this.b*(first+1)-1; // last cell of the starting block
			int startLast=last*this.b; // first cell of the ending block
			
			result=rmqB.getMin(first+1, last-1);
			result=Data.min(result, rmqT.getMin(l, endFirst));
			result=Data.min(result, rmqT.getMin(startLast, r));
		}
		
		return result;
	}

	public String toString() {
		return "efficient implementation: n="+this.table.length+", block size="+this.b;
	}

}
