import java.util.LinkedList;
import java.util.Queue;
import java.util.Vector;

/** Classe qui représente un (noeud d'un) arbre général: le nombre de fils est arbitraire. <br>
 * L'arbre vide est représenté par 'null'. */
public class Tree {

	/** donnée stoquée dans chaque noeud */
	String element;

	/** tableau dynamique qui stoque les fils du noeud courant */
	Vector<Tree> children;

	/** 
	 * Construit un noeud de l'arbre arbre (avec par étiquette 'element'). 
	 */
	Tree(String element) {
		this.element=element;
		this.children=new Vector<Tree>();
	}

	/** Renvoie le LCA des deux noeuds ayant pour étiquettes 'el1' and 'el2'. <br>
	 * 
	 * @param tree  noeud racine de l'arbre
	 */
	static Tree lca(Tree node, String el1, String el2) {
		//throw new Error("\nMéthode size() à compléter (Question 2.1)");
		if (node==null)
			return null;

		if (node.element.equals(el1) || node.element.equals(el2)) // an element is ancestor of the other one
			return node;

		Vector<Tree> subtrees=new Vector<Tree>(); // for storing the result of recursive calls into sub-trees

		for(Tree child: node.children) { // iterate over the sub-trees of the current node
			Tree query=lca(child, el1, el2);
			if(query!=null)
				subtrees.add(query);
		}
		if(subtrees.size()>2) // only for debugging
			throw new Error("Error: wrong number of sub-trees... only at most two sub-trees are expected to contains the query elements");

		if(subtrees.size()==0) // no sub-tree contain 'el1' or 'el2'
			return null;
		else if (subtrees.size()==2) // we found the LCA of el1 and el2
			return node;
		else
			return subtrees.get(0);
	}

	/** Renvoie le nombre de noeuds de l'arbre */
	static int size(Tree node) {
		//throw new Error("\nMéthode size() à compléter (Question 2.2)");
		if (node==null)
			return 0;
		else {
			int result=1;
			for(Tree child: node.children) {
				result=result+size(child);
			}
			return result;
		}
	}

	/** Affiche les noeuds de l'arbre selon un parcours BFS (par niveaux)
	 * @param node  noeud racine de l'arbre
	 */
	static void print(Tree node) {
		if (node == null)
			return;

		Queue<Tree> nodes = new LinkedList<Tree>();
		nodes.add(node);
		while(nodes.isEmpty()==false) {
			int n = nodes.size();

			while (n > 0) {
				Tree p = nodes.peek();
				nodes.remove();
				System.out.print(p.element + " ");

				for(Tree child: p.children)
					nodes.add(child);
				n--;
			}

			System.out.println(); 
		}
	}

}
